@@ -31,7 +31,7 @@ except ImportError:
     pep8 = None
 
 
-__version__ = '0.7'
+__version__ = '0.7.1'
 
 
 PEP8_BIN = 'pep8'
@@ -59,9 +59,10 @@ def detect_encoding(filename):
             with open(filename, 'rb') as input_file:
                 encoding = tokenize.detect_encoding(input_file.readline)[0]
 
-            # Check for correctness of encoding
-            with open(filename, encoding=encoding) as input_file:
-                input_file.read()
+                # Check for correctness of encoding
+                import io
+                with io.TextIOWrapper(input_file, encoding) as wrapper:
+                    wrapper.read()
 
             return encoding
         except (SyntaxError, LookupError, UnicodeDecodeError):
@@ -108,9 +109,11 @@ class FixPEP8(object):
             sio = StringIO(contents)
             self.source = sio.readlines()
         self.original_source = copy.copy(self.source)
-        self.newline = _find_newline(self.source)
+        self.newline = find_newline(self.source)
         self.options = options
         self.indent_word = _get_indentword("".join(self.source))
+        self.logical_start = None
+        self.logical_end = None
         # method definition
         self.fix_e111 = self.fix_e101
         self.fix_e202 = self.fix_e201
@@ -129,7 +132,19 @@ class FixPEP8(object):
 
     def _fix_source(self, results):
         completed_lines = []
+        global_read_fixes = ['e12']
         for result in sorted(results, key=_priority_key):
+            # Some fixes read all lines and thus will break if other fixes
+            # modify the lines.
+            is_global_read_fix = False
+            for prefix in global_read_fixes:
+                if result['id'].lower().startswith(prefix):
+                    is_global_read_fix = True
+
+            # Do not run global fix if any lines have been modified.
+            if is_global_read_fix and completed_lines:
+                continue
+
             if result['line'] in completed_lines:
                 continue
             fixed_methodname = "fix_%s" % result['id'].lower()
@@ -165,6 +180,9 @@ class FixPEP8(object):
             }
             results = _execute_pep8(pep8_options, self.source)
         else:
+            if self.options.verbose:
+                sys.stderr.write('Running in compatibility mode. Consider '
+                                 'upgrading to the latest pep8.\n')
             results = _spawn_pep8((["--ignore=" + self.options.ignore]
                                    if self.options.ignore else []) +
                                   (["--select=" + self.options.select]
@@ -183,6 +201,145 @@ class FixPEP8(object):
         else:
             return []
 
+    def find_logical(self, force=False):
+        # make a variable which is the index of all the starts of lines
+        if force is False and self.logical_start is not None:
+            return
+        logical_start = []
+        logical_end = []
+        last_newline = True
+        sio = StringIO("".join(self.source))
+        parens = 0
+        for t in tokenize.generate_tokens(sio.readline):
+            if t[0] in [tokenize.COMMENT, tokenize.DEDENT,
+                        tokenize.INDENT, tokenize.NL,
+                        tokenize.ENDMARKER]:
+                continue
+            if not parens and t[0] in [
+                tokenize.NEWLINE, tokenize.SEMI
+            ]:
+                last_newline = True
+                logical_end.append((t[3][0] - 1, t[2][1]))
+                continue
+            if last_newline and not parens:
+                logical_start.append((t[2][0] - 1, t[2][1]))
+                last_newline = False
+            if t[0] == tokenize.OP:
+                if t[1] in '([{':
+                    parens += 1
+                elif t[1] in '}])':
+                    parens -= 1
+        self.logical_start = logical_start
+        self.logical_end = logical_end
+
+    def get_logical(self, result):
+        """return the logical line corresponding to the result.  Assumes input
+        is already E702-clean"""
+        self.find_logical()
+        row = result['line'] - 1
+        col = result['column'] - 1
+        ls = None
+        le = None
+        for i in range(0, len(self.logical_start), 1):
+            x = self.logical_end[i]
+            if x[0] > row or (x[0] == row and x[1] > col):
+                le = x
+                ls = self.logical_start[i]
+                break
+        if ls is None:
+            return None
+        original = self.source[ls[0]:le[0] + 1]
+        return ls, le, original
+
+    def _fix_reindent(self, result, fix_distinct=False):
+        """Fix a badly indented line by adding or removing from its initial
+        indent only."""
+        logical = self.get_logical(result)
+        if not logical:
+            return []
+        ls, _, original = logical
+        try:
+            rewrapper = Wrapper(original, hard_wrap=MAX_LINE_WIDTH)
+        except (tokenize.TokenError, IndentationError):
+            return []
+        valid_indents = rewrapper.pep8_expected()
+        if not rewrapper.rel_indent:
+            return []
+        if result["line"] > ls[0]:
+            # got a valid continuation line number from pep8
+            row = result["line"] - ls[0] - 1
+            # always pick the first option for this
+            valid = valid_indents[row]
+            got = rewrapper.rel_indent[row]
+        else:
+            # line number from pep8 isn't a continuation line.  instead,
+            # compare our own function's result, look for the first mismatch,
+            # and just hope that we take fewer than 100 iterations to finish.
+            for row in range(0, len(original), 1):
+                valid = valid_indents[row]
+                got = rewrapper.rel_indent[row]
+                if valid != got:
+                    break
+        line = ls[0] + row
+        orig_line = self.source[line]
+        # always pick the expected indent, for now.
+        indent_to = valid[0]
+        if fix_distinct and indent_to == 4:
+            if len(valid) == 1:
+                return []
+            else:
+                indent_to = valid[1]
+
+        if got != indent_to:
+            self.source[line] = (
+                " " * (indent_to) + orig_line.lstrip()
+            )
+        else:
+            return []
+
+    def fix_e121(self, result):
+        """The 'peculiar indent' error for hanging indents"""
+        # fix by adjusting initial indent level
+        return self._fix_reindent(result)
+
+    def fix_e122(self, result):
+        """The 'absent indent' error for hanging indents"""
+        # fix by adding an initial indent
+        return self._fix_reindent(result)
+
+    def fix_e123(self, result):
+        """The 'loose fingernails' indentation level error for hanging
+        indents"""
+        # fix by deleting whitespace to the correct level
+        return self._fix_reindent(result)
+
+    def fix_e124(self, result):
+        """The 'loose fingernails' indentation level error for visual
+        indents"""
+        # fix by inserting whitespace before the closing bracket
+        return self._fix_reindent(result)
+
+    def fix_e125(self, result):
+        """The 'often not visually distinct' error."""
+        # fix by indenting the line in error to the next stop.
+        return self._fix_reindent(result, fix_distinct=True)
+
+    def fix_e126(self, result):
+        """The 'spectacular indent' error for hanging indents"""
+        # fix by deleting whitespace to the left
+        return self._fix_reindent(result)
+
+    def fix_e127(self, result):
+        """The 'interpretive dance' indentation error."""
+        # fix by deleting whitespace to the correct level
+        return self._fix_reindent(result)
+
+    def fix_e128(self, result):
+        """The 'I just wrap long lines with a line break in the middle of my
+        argument list' indentation error."""
+        # fix by deleting whitespace to the correct level
+        return self._fix_reindent(result)
+
     def fix_e201(self, result):
         line_index = result['line'] - 1
         target = self.source[line_index]
@@ -224,7 +381,10 @@ class FixPEP8(object):
         line_index = result['line'] - 1
         target = self.source[line_index]
         c = result['column'] - 1
-        fixed = target[:c] + re.sub(r'\s*=\s*', '=', target[c:], 1)
+        if target[c].strip():
+            fixed = target
+        else:
+            fixed = target[:c].rstrip() + target[c:].lstrip()
 
         # There could be an escaped newline
         #
@@ -301,6 +461,9 @@ class FixPEP8(object):
         line_index = result['line'] - 1
         target = self.source[line_index]
 
+        if not target.lstrip().startswith('import'):
+            return []
+
         indentation = target.split("import ")[0]
         modules = target.split("import ")[1].split(",")
         fixed_modulelist = \
@@ -336,11 +499,11 @@ class FixPEP8(object):
             # my_long_function_name(x, y,
             #     z, ...)
             candidate0 = _shorten_line(_tokens, source, target, indent,
-                                       self.indent_word, reverse=False,
-                                       newline=self.newline)
+                                       self.indent_word, newline=self.newline,
+                                       reverse=False)
             candidate1 = _shorten_line(_tokens, source, target, indent,
-                                       self.indent_word, reverse=True,
-                                       newline=self.newline)
+                                       self.indent_word, newline=self.newline,
+                                       reverse=True)
             if candidate0 and candidate1:
                 if candidate0.split(self.newline)[0].endswith('('):
                     self.source[line_index] = candidate0
@@ -585,7 +748,7 @@ class FixPEP8(object):
 
     def fix_w603(self, result):
         target = self.source[result['line'] - 1]
-        self.source[result['line'] - 1] = re.sub('<>', '!=', target)
+        self.source[result['line'] - 1] = re.sub('<>', '!=', target, count=1)
 
     def fix_w604(self, result):
         target = self.source[result['line'] - 1]
@@ -604,7 +767,7 @@ class FixPEP8(object):
             target[:start], target[start + 1:end - 1], target[end:])
 
 
-def _find_newline(source):
+def find_newline(source):
     """Return type of newline used in source."""
     cr, lf, crlf = 0, 0, 0
     for s in source:
@@ -676,7 +839,10 @@ def _priority_key(pep8_result):
     """
     priority = ['e101', 'e111', 'w191',  # Global fixes
                 'e701',  # Fix multiline colon-based before semicolon based
-                'e702']  # Break multiline statements early
+                'e702',  # Break multiline statements early
+                'e225', 'e231',  # things that make lines longer
+                'e501',  # before we break lines
+                ]
     key = pep8_result['id'].lower()
     if key in priority:
         return priority.index(key)
@@ -724,8 +890,8 @@ def _fix_basic_raise(line, newline):
                     comment, newline])
 
 
-def _shorten_line(tokens, source, target, indentation, indent_word,
-                  reverse=False, newline=LF):
+def _shorten_line(tokens, source, target, indentation, indent_word, newline,
+                  reverse=False):
     """Separate line at OPERATOR."""
     if reverse:
         tokens.reverse()
@@ -744,13 +910,19 @@ def _shorten_line(tokens, source, target, indentation, indent_word,
                     continue
             first = source[:offset - len(indentation)]
             second = (indentation + indent_word +
-                      source[offset - len(indentation):])
+                      source[offset - len(indentation):].lstrip())
             if not second.strip():
                 continue
             # Don't modify if lines are not short enough
             if len(first) > MAX_LINE_WIDTH or len(second) > MAX_LINE_WIDTH:
                 continue
-            fixed = first + newline + second
+            # Do not begin a line with a comma
+            if second.lstrip().startswith(','):
+                continue
+            if tkn[1] in '+-*/':
+                fixed = first + ' \\' + newline + second
+            else:
+                fixed = first + newline + second
             try:
                 ret = compile(fixed, '<string>', 'exec')
             except SyntaxError:
@@ -912,9 +1084,8 @@ class Reindenter(object):
                         for j in range(i - 1, -1, -1):
                             jline, jlevel = stats[j]
                             if jlevel >= 0:
-                                want = (have +
-                                        _leading_space_count(
-                                            after[jline - 1]) -
+                                want = (have + _leading_space_count(
+                                        after[jline - 1]) -
                                         _leading_space_count(lines[jline]))
                                 break
                     if want < 0:
@@ -951,35 +1122,30 @@ class Reindenter(object):
             self.index += 1
         return line
 
-    def tokeneater(self, token_type, _, start, __, line,
-                   INDENT=tokenize.INDENT,
-                   DEDENT=tokenize.DEDENT,
-                   NEWLINE=tokenize.NEWLINE,
-                   COMMENT=tokenize.COMMENT,
-                   NL=tokenize.NL):
+    def tokeneater(self, token_type, _, start, __, line):
         """Line-eater for tokenize."""
         sline = start[0]
-        if token_type == NEWLINE:
+        if token_type == tokenize.NEWLINE:
             # A program statement, or ENDMARKER, will eventually follow,
             # after some (possibly empty) run of tokens of the form
             #     (NL | COMMENT)* (INDENT | DEDENT+)?
             self.find_stmt = 1
 
-        elif token_type == INDENT:
+        elif token_type == tokenize.INDENT:
             self.find_stmt = 1
             self.level += 1
 
-        elif token_type == DEDENT:
+        elif token_type == tokenize.DEDENT:
             self.find_stmt = 1
             self.level -= 1
 
-        elif token_type == COMMENT:
+        elif token_type == tokenize.COMMENT:
             if self.find_stmt:
                 self.stats.append((sline, -1))
                 # but we're still looking for a new stmt, so leave
                 # find_stmt alone
 
-        elif token_type == NL:
+        elif token_type == tokenize.NL:
             pass
 
         elif self.find_stmt:
@@ -991,6 +1157,262 @@ class Reindenter(object):
                 self.stats.append((sline, self.level))
 
 
+class Wrapper(object):
+
+    """Class for functions relating to continuation lines and line folding.
+
+    Each instance operates on a single logical line.
+
+    """
+
+    SKIP_TOKENS = frozenset([
+        tokenize.COMMENT, tokenize.NL, tokenize.INDENT,
+        tokenize.DEDENT, tokenize.NEWLINE, tokenize.ENDMARKER
+    ])
+
+    def __init__(self, physical_lines, hard_wrap=79, soft_wrap=72):
+        if type(physical_lines) != list:
+            physical_lines = physical_lines.splitlines(keepends=True)
+        self.lines = physical_lines
+        self.index = 0
+        self.hard_wrap = hard_wrap
+        self.soft_wrap = soft_wrap
+        self.tokens = list()
+        self.rel_indent = None
+        sio = StringIO("".join(physical_lines))
+        for t in tokenize.generate_tokens(sio.readline):
+            if t[0] != tokenize.ENDMARKER:
+                #if t[2][0] > max_seen:
+                    #max_seen = t[2][0]
+                    #print ">>" + repr(t[4]) + "<<"
+                self.tokens.append(t)
+        self.logical_line, self.mapping = self.build_tokens_logical(
+            self.tokens
+        )
+
+    def build_tokens_logical(self, tokens):
+        """
+        Build a logical line from a list of tokens.  Returns the logical line
+        and a list of (offset, token) tuples.  Does not mute strings like the
+        version in pep8.py.
+        """
+        # from pep8.py with minor modifications
+        mapping = []
+        logical = []
+        length = 0
+        previous = None
+        for t in tokens:
+            token_type, text = t[0:2]
+            if token_type in self.SKIP_TOKENS:
+                continue
+            if previous:
+                end_line, end = previous[3]
+                start_line, start = t[2]
+                if end_line != start_line:  # different row
+                    prev_text = self.lines[end_line - 1][end - 1]
+                    if prev_text == ',' or (prev_text not in '{[('
+                                            and text not in '}])'):
+                        logical.append(' ')
+                        length += 1
+                elif end != start:  # different column
+                    fill = self.lines[end_line - 1][end:start]
+                    logical.append(fill)
+                    length += len(fill)
+            mapping.append((length, t))
+            logical.append(text)
+            length += len(text)
+            previous = t
+        logical_line = ''.join(logical)
+        assert logical_line.lstrip() == logical_line
+        assert logical_line.rstrip() == logical_line
+        return logical_line, mapping
+
+    def pep8_expected(self):
+        """Helper function which replicates logic in pep8.py, to know what
+        level to indent things to.
+
+        Returns a list of lists; each list represents valid indent levels for
+        the line in question, relative from the initial indent.  However, the
+        first entry is the indent level which was expected.
+        """
+
+        # what follows is an adjusted version of
+        # pep8.py:continuation_line_indentation.  All of the comments have been
+        # stripped and the 'yield' statements replaced with 'pass'.
+        tokens = self.tokens
+
+        first_row = tokens[0][2][0]
+        nrows = 1 + tokens[-1][2][0] - first_row
+
+        # here are the return values
+        valid_indents = [list()] * nrows
+        indent_level = tokens[0][2][1]
+        valid_indents[0].append(indent_level)
+
+        if nrows == 1:
+            # bug, really.
+            return valid_indents
+
+        indent_next = self.logical_line.endswith(':')
+
+        row = depth = 0
+        parens = [0] * nrows
+        self.rel_indent = rel_indent = [0] * nrows
+        indent = [indent_level]
+        indent_chances = {}
+        last_indent = (0, 0)
+        last_token_multiline = None
+
+        for token_type, text, start, end, _ in self.tokens:
+            newline = row < start[0] - first_row
+            if newline:
+                row = start[0] - first_row
+                newline = (not last_token_multiline and
+                           token_type not in (tokenize.NL, tokenize.NEWLINE))
+
+            if newline:
+                # this is where the differences start.  instead of looking at
+                # the line and determining whether the observed indent matches
+                # our expectations, we decide which type of indentation is in
+                # use at the given indent level, and return the offset.  this
+                # algorithm is succeptable to "carried errors", but should
+                # through repeated runs eventually solve indentation for
+                # multi-line expressions less than PEP8_PASSES_MAX lines long.
+
+                if depth:
+                    for open_row in range(row - 1, -1, -1):
+                        if parens[open_row]:
+                            break
+                else:
+                    open_row = 0
+
+                # that's all we get to work with.  this code attempts to
+                # "reverse" the below logic, and place into the valid indents
+                # list
+                vi = []
+                add_second_chances = False
+                if token_type == tokenize.OP and text in ']})':
+                    # this line starts with a closing bracket, so it needs to
+                    # be closed at the same indent as the opening one.
+                    if indent[depth]:
+                        # hanging indent
+                        vi.append(indent[depth])
+                    else:
+                        # visual indent
+                        vi.append(rel_indent[open_row])
+                elif depth and indent[depth]:
+                    # visual indent was previously confirmed.
+                    vi.append(indent[depth])
+                    add_second_chances = True
+                elif depth and True in indent_chances.values():
+                    # visual indent happened before, so stick to
+                    # visual indent this time.
+                    if depth > 1 and indent[depth - 1]:
+                        vi.append(indent[depth - 1])
+                    else:
+                        # stupid fallback
+                        vi.append(indent_level + 4)
+                    add_second_chances = True
+                elif not depth:
+                    vi.append(indent_level + 4)
+                else:
+                    # must be in hanging indent
+                    hang = rel_indent[open_row] + 4
+                    vi.append(indent_level + hang)
+
+                # about the best we can do without look-ahead
+                if indent_next and vi[0] == indent_level + 4 and \
+                        nrows == row + 1:
+                    vi[0] += 4
+
+                if add_second_chances:
+                    # visual indenters like to line things up.
+                    min_indent = vi[0]
+                    for col, what in indent_chances.items():
+                        if col > min_indent and (
+                            what is True or
+                            (what == str and token_type == tokenize.STRING) or
+                            (what == text and token_type == tokenize.OP)
+                        ):
+                            vi.append(col)
+                    vi = sorted(vi)
+
+                valid_indents[row] = vi
+
+                # ...returning to original continuation_line_identation func...
+                visual_indent = indent_chances.get(start[1])
+                last_indent = start
+                rel_indent[row] = start[1] - indent_level
+                hang = rel_indent[row] - rel_indent[open_row]
+
+                if token_type == tokenize.OP and text in ']})':
+                    if indent[depth]:
+                        if start[1] != indent[depth]:
+                            pass  # E124
+                    elif hang:
+                        pass  # E123
+                elif visual_indent is True:
+                    if not indent[depth]:
+                        indent[depth] = start[1]
+                elif visual_indent in (text, str):
+                    pass
+                elif indent[depth] and start[1] < indent[depth]:
+                    pass  # E128
+                elif hang == 4 or (indent_next and rel_indent[row] == 8):
+                    pass
+                else:
+                    if hang <= 0:
+                        pass  # E122
+                    elif indent[depth]:
+                        pass  # E127
+                    elif hang % 4:
+                        pass  # E121
+                    else:
+                        pass  # E126
+
+            # line altered: comments shouldn't define a visual indent
+            if parens[row] and not indent[depth] and token_type not in (
+                tokenize.NL, tokenize.COMMENT
+            ):
+                indent[depth] = start[1]
+                indent_chances[start[1]] = True
+            elif token_type == tokenize.STRING or text in (
+                'u', 'ur', 'b', 'br'
+            ):
+                indent_chances[start[1]] = str
+
+            if token_type == tokenize.OP:
+                if text in '([{':
+                    depth += 1
+                    indent.append(0)
+                    parens[row] += 1
+                elif text in ')]}' and depth > 0:
+                    prev_indent = indent.pop() or last_indent[1]
+                    for d in range(depth):
+                        if indent[d] > prev_indent:
+                            indent[d] = 0
+                    for ind in list(indent_chances):
+                        if ind >= prev_indent:
+                            del indent_chances[ind]
+                    depth -= 1
+                    if depth:
+                        indent_chances[indent[depth]] = True
+                    for idx in range(row, -1, -1):
+                        if parens[idx]:
+                            parens[idx] -= 1
+                            break
+                assert len(indent) == depth + 1
+                if start[1] not in indent_chances:
+                    indent_chances[start[1]] = text
+
+            last_token_multiline = (start[0] != end[0])
+
+        if indent_next and rel_indent[-1] == 4:
+            pass  # E125
+
+        return valid_indents
+
+
 def _leading_space_count(line):
     """Return number of leading spaces in line."""
     i = 0
@@ -1001,6 +1423,12 @@ def _leading_space_count(line):
 
 def fix_file(filename, opts, output=sys.stdout):
     tmp_source = read_from_filename(filename)
+
+    # Add missing newline (important for diff)
+    tmp_newline = find_newline(tmp_source)
+    if tmp_source == tmp_source.rstrip(tmp_newline):
+        tmp_source += tmp_newline
+
     fix = FixPEP8(filename, opts, contents=tmp_source)
     fixed_source = fix.fix()
     original_source = copy.copy(fix.original_source)
@@ -1049,6 +1477,9 @@ def parse_args(args):
                       help='print the diff for the fixed source')
     parser.add_option('-i', '--in-place', action='store_true',
                       help='make changes to files in place')
+    parser.add_option('-r', '--recursive', action='store_true',
+                      help='run recursively; must be used with --in-place or '
+                           '--diff')
     parser.add_option('-p', '--pep8-passes',
                       default=PEP8_PASSES_MAX, type='int',
                       help='maximum number of additional pep8 passes'
@@ -1067,23 +1498,38 @@ def parse_args(args):
                      'unless the "--in-place" or "--diff" options are '
                      'used')
 
+    if opts.recursive and not (opts.in_place or opts.diff):
+        parser.error('must be used with --in-place or --diff')
+
+    if opts.in_place and opts.diff:
+        parser.error('--in-place and --diff are mutually exclusive')
+
     return opts, args
 
 
 def main():
     """Tool main."""
     opts, args = parse_args(sys.argv[1:])
-    try:
-        if opts.in_place or opts.diff:
-            for f in set(args):
-                if opts.verbose and len(args) > 1:
-                    sys.stderr.write('[file:%s]\n' % f)
-                fix_file(f, opts)
+    if opts.in_place or opts.diff:
+        filenames = list(set(args))
+    else:
+        assert len(args) == 1
+        assert not opts.recursive
+        filenames = args[:1]
+
+    while filenames:
+        name = filenames.pop(0)
+        if opts.recursive and os.path.isdir(name):
+            for root, _, children in os.walk(name):
+                filenames += [os.path.join(root, f) for f in children
+                              if f.endswith('.py')]
         else:
-            fix_file(args[0], opts)
-    except IOError as error:
-        sys.stderr.write(str(error) + '\n')
-        return 1
+            if opts.verbose:
+                sys.stderr.write('[file:%s]\n' % name)
+            try:
+                fix_file(name, opts)
+            except (UnicodeDecodeError, IOError) as error:
+                sys.stderr.write(str(error) + '\n')
 
 
 if __name__ == '__main__':
