@@ -1,6 +1,7 @@
 import os
 import sys
 import unittest
+import contextlib
 from subprocess import Popen, PIPE
 from tempfile import mkstemp
 try:
@@ -42,23 +43,23 @@ class TestUtils(unittest.TestCase):
 
     def test_find_newline_only_cr(self):
         source = ["print 1\r", "print 2\r", "print3\r"]
-        self.assertEqual(autopep8.CR, autopep8._find_newline(source))
+        self.assertEqual(autopep8.CR, autopep8.find_newline(source))
 
     def test_find_newline_only_lf(self):
         source = ["print 1\n", "print 2\n", "print3\n"]
-        self.assertEqual(autopep8.LF, autopep8._find_newline(source))
+        self.assertEqual(autopep8.LF, autopep8.find_newline(source))
 
     def test_find_newline_only_crlf(self):
         source = ["print 1\r\n", "print 2\r\n", "print3\r\n"]
-        self.assertEqual(autopep8.CRLF, autopep8._find_newline(source))
+        self.assertEqual(autopep8.CRLF, autopep8.find_newline(source))
 
     def test_find_newline_cr1_and_lf2(self):
         source = ["print 1\n", "print 2\r", "print3\n"]
-        self.assertEqual(autopep8.LF, autopep8._find_newline(source))
+        self.assertEqual(autopep8.LF, autopep8.find_newline(source))
 
     def test_find_newline_cr1_and_crlf2(self):
         source = ["print 1\r\n", "print 2\r", "print3\r\n"]
-        self.assertEqual(autopep8.CRLF, autopep8._find_newline(source))
+        self.assertEqual(autopep8.CRLF, autopep8.find_newline(source))
 
     def test_shorten_comment(self):
         self.assertEqual('# ' + 'x' * 77 + '\n',
@@ -103,9 +104,8 @@ class TestFixPEP8Error(unittest.TestCase):
         os.remove(self.tempfile[1])
 
     def _inner_setup(self, line, options=""):
-        f = open(self.tempfile[1], 'w')
-        f.write(line)
-        f.close()
+        with open(self.tempfile[1], 'w') as temp_file:
+            temp_file.write(line)
         opts, _ = autopep8.parse_args(options.split() + [self.tempfile[1]])
         sio = StringIO()
         autopep8.fix_file(filename=self.tempfile[1],
@@ -250,6 +250,39 @@ def foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e12_reindent(self):
+        line = """
+
+def foo_bar(baz, frop,
+    fizz, bang):
+    pass
+"""
+        fixed = """
+
+def foo_bar(baz, frop,
+            fizz, bang):
+    pass
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e12_reindent_with_multiple_fixes(self):
+        line="""
+
+sql = 'update %s set %s %s' % (from_table,
+                               ','.join(['%s=%s' % (col, col) for col in cols]),
+        where_clause)
+"""
+        fixed = """
+
+sql = 'update %s set %s %s' % (from_table,
+                               ','.join(
+                                   ['%s=%s' % (col, col) for col in cols]),
+                               where_clause)
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e191(self):
         line = """
 while True:
@@ -432,6 +465,12 @@ class Foo(object):
         self._inner_setup(line, options='--select=E251')
         self.assertEqual(self.result, fixed)
 
+    def test_e251_with_calling(self):
+        line = "foo(bar= True)\n"
+        fixed = "foo(bar=True)\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e261(self):
         line = "print 'a b '# comment\n"
         fixed = "print 'a b '  # comment\n"
@@ -440,13 +479,13 @@ class Foo(object):
 
     def test_e261_with_dictionary(self):
         line = "d = {# comment\n1: 2}\n"
-        fixed = "d = {  # comment\n1: 2}\n"
+        fixed = "d = {  # comment\n    1: 2}\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
     def test_e261_with_dictionary_no_space(self):
         line = "d = {#comment\n1: 2}\n"
-        fixed = "d = {  # comment\n1: 2}\n"
+        fixed = "d = {  # comment\n    1: 2}\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
@@ -564,7 +603,7 @@ class Foo(object):
 """
         fixed = \
 """print(111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222, 222, 333,
-     333, 333, 333)
+      333, 333, 333)
 """
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
@@ -577,7 +616,20 @@ class Foo(object):
         fixed = \
 """def d():
     print(111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222,
-         222, 333, 333, 333, 333)
+          222, 333, 333, 333, 333)
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e501_arithmetic_operator_with_indent(self):
+        line = \
+"""def d():
+    111 + 111 + 111 + 111 + 111 + 222 + 222 + 222 + 222 + 222 + 222 + 222 + 222 + 222 + 333 + 333 + 333 + 333
+"""
+        fixed = \
+"""def d():
+    111 + 111 + 111 + 111 + 111 + 222 + 222 + 222 + 222 + 222 + 222 + \\
+        222 + 222 + 222 + 333 + 333 + 333 + 333
 """
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
@@ -632,6 +684,30 @@ class Foo(object):
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e501_do_not_begin_line_with_comma(self):
+        # This fix is incomplete. (The line is still too long.) But it is here
+        # just to confirm that we do not put a comma at the beginning of a
+        # line.
+        line = """
+
+def dummy():
+    if True:
+        if True:
+            if True:
+                object = ModifyAction( [MODIFY70.text, OBJECTBINDING71.text, COLON72.text], MODIFY70.getLine(), MODIFY70.getCharPositionInLine() )
+"""
+        fixed = """
+
+def dummy():
+    if True:
+        if True:
+            if True:
+                object = ModifyAction([MODIFY70.text, OBJECTBINDING71.text, COLON72.text],
+                                      MODIFY70.getLine(), MODIFY70.getCharPositionInLine())
+"""
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e501_do_not_break_if_useless(self):
         line = \
 """
@@ -678,8 +754,8 @@ class Foo(object):
         self.assertEqual(self.result, fixed)
 
     def test_e502(self):
-        line = "print('abc'\\\n'def')\n"
-        fixed = "print('abc'\n'def')\n"
+        line = "print('abc'\\\n      'def')\n"
+        fixed = "print('abc'\n      'def')\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
@@ -752,9 +828,8 @@ class TestFixPEP8Warning(unittest.TestCase):
         os.remove(self.tempfile[1])
 
     def _inner_setup(self, line):
-        f = open(self.tempfile[1], 'w')
-        f.write(line)
-        f.close()
+        with open(self.tempfile[1], 'w') as temp_file:
+            temp_file.write(line)
         opts, _ = autopep8.parse_args([self.tempfile[1]])
         sio = StringIO()
         autopep8.fix_file(filename=self.tempfile[1],
@@ -856,7 +931,8 @@ class TestFixPEP8Warning(unittest.TestCase):
         # We don't support this case
         line = """
 a.has_key(
-0)
+    0
+)
 """.lstrip()
         fixed = line
         self._inner_setup(line)
@@ -905,7 +981,7 @@ a.has_key(
     @only_py2
     def test_w602_skip_complex_multiline(self):
         # We do not handle formatted multiline strings
-        line = 'raise ValueError, """\nhello %s %s""" % (1,\n2)\n'
+        line = 'raise ValueError, """\nhello %s %s""" % (\n    1, 2)\n'
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
@@ -1042,7 +1118,7 @@ a.has_key(
 
     def test_w604_skip_multiple_lines(self):
         # We do not support this yet
-        line = "`(1\n)`\n"
+        line = "`(1\n  )`\n"
         fixed = line
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
@@ -1057,9 +1133,8 @@ class TestOptions(unittest.TestCase):
         os.remove(self.tempfile[1])
 
     def _inner_setup(self, line, options):
-        f = open(self.tempfile[1], 'w')
-        f.write(line)
-        f.close()
+        with open(self.tempfile[1], 'w') as temp_file:
+            temp_file.write(line)
         p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1]] + options,
                   stdout=PIPE)
         self.result = p.communicate()[0].decode('utf8')
@@ -1098,6 +1173,18 @@ class TestOptions(unittest.TestCase):
         verbose_error = p.communicate()[1].decode('utf8')
         self.assertIn("'fix_e901' is not defined", verbose_error)
 
+    @py27_and_above
+    def test_verbose_diff(self):
+        line = 'bad_syntax)'
+        f = open(self.tempfile[1], 'w')
+        f.write(line)
+        f.close()
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) +
+                  [self.tempfile[1], '--verbose', '--diff'],
+                  stdout=PIPE, stderr=PIPE)
+        verbose_error = p.communicate()[1].decode('utf8')
+        self.assertIn("'fix_e901' is not defined", verbose_error)
+
     def test_in_place(self):
         line = "'abc'  \n"
         fixed = "'abc'\n"
@@ -1112,6 +1199,56 @@ class TestOptions(unittest.TestCase):
         self.assertEqual(f.read(), fixed)
         f.close()
 
+    def test_in_place_and_diff(self):
+        line = "'abc'  \n"
+        fixed = "'abc'\n"
+        f = open(self.tempfile[1], 'w')
+        f.write(line)
+        f.close()
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--in-place', '--diff'],
+                  stderr=PIPE)
+        result = p.communicate()[1].decode('utf8')
+        self.assertTrue('--in-place and --diff are mutually exclusive' in result)
+
+    def test_recursive(self):
+        import tempfile
+        temp_directory = tempfile.mkdtemp(dir='.')
+        try:
+            with open(os.path.join(temp_directory, 'a.py'), 'w') as output:
+                output.write("'abc'  \n")
+
+            os.mkdir(os.path.join(temp_directory, 'd'))
+            with open(os.path.join(temp_directory, 'd', 'b.py'),
+                      'w') as output:
+                output.write("123  \n")
+
+            p = Popen(list(AUTOPEP8_CMD_TUPLE) +
+                      [temp_directory, '--recursive', '--diff'],
+                      stdout=PIPE)
+            result = p.communicate()[0].decode('utf8')
+
+            self.assertEqual(
+                "-'abc'  \n+'abc'",
+                '\n'.join(result.split('\n')[3:5]))
+
+            self.assertEqual(
+                '-123  \n+123',
+                '\n'.join(result.split('\n')[8:10]))
+        finally:
+            import shutil
+            shutil.rmtree(temp_directory)
+
+    def test_only_recursive(self):
+        line = "'abc'  \n"
+        fixed = "'abc'\n"
+        f = open(self.tempfile[1], 'w')
+        f.write(line)
+        f.close()
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--recursive'],
+                  stderr=PIPE)
+        result = p.communicate()[1].decode('utf8')
+        self.assertTrue('must be used with --in-place or --diff' in result)
+
 
 class TestSpawnPEP8Process(unittest.TestCase):
 
@@ -1122,9 +1259,8 @@ class TestSpawnPEP8Process(unittest.TestCase):
         os.remove(self.tempfile[1])
 
     def _inner_setup(self, line, options=""):
-        f = open(self.tempfile[1], 'w')
-        f.write(line)
-        f.close()
+        with open(self.tempfile[1], 'w') as temp_file:
+            temp_file.write(line)
         opts, _ = autopep8.parse_args(options.split() + [self.tempfile[1]])
         sio = StringIO()
 
@@ -1147,5 +1283,51 @@ class TestSpawnPEP8Process(unittest.TestCase):
         self.assertEqual(self.result, fixed)
 
 
+class TestCoverage(unittest.TestCase):
+
+    def setUp(self):
+        self.tempfile = mkstemp()
+
+    def tearDown(self):
+        os.remove(self.tempfile[1])
+
+    def _inner_setup(self, line, options=""):
+        with open(self.tempfile[1], 'w') as temp_file:
+            temp_file.write(line)
+
+    def test_fixpep8_class_constructor(self):
+        line = "print 1\nprint 2\n"
+        self._inner_setup(line)
+        pep8obj = autopep8.FixPEP8(self.tempfile[1], None)
+        self.assertEqual("".join(pep8obj.source), line)
+
+    def test_no_argument(self):
+        with disable_stderr():
+            try:
+                autopep8.parse_args([])
+                self.assertEqual("not work", "test has failed!!")
+            except SystemExit as e:
+                self.assertEqual(e.code, 2)
+
+    def test_inplace_with_multi_files(self):
+        with disable_stderr():
+            try:
+                autopep8.parse_args(['test.py', 'dummy.py'])
+                self.assertEqual("not work", "test has failed!!")
+            except SystemExit as e:
+                self.assertEqual(e.code, 2)
+
+
+@contextlib.contextmanager
+def disable_stderr():
+    with open('/dev/null', 'w') as fake_stderr:
+        _tmp = sys.stderr
+        sys.stderr = fake_stderr
+        try:
+            yield
+        finally:
+            sys.stderr = _tmp
+
+
 if __name__ == '__main__':
     unittest.main()
